#!@l_prefix@/bin/perl
##
##  registry-ui.pl - OpenPKG registration user interface (mod_perl)
##  Copyright (c) 2000-2005 OpenPKG Foundation e.V. <http://openpkg.net/>
##  Copyright (c) 2000-2005 Ralf S. Engelschall <http://engelschall.com/>
##
##  Permission to use, copy, modify, and distribute this software for
##  any purpose with or without fee is hereby granted, provided that
##  the above copyright notice and this permission notice appear in all
##  copies.
##
##  THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED
##  WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
##  MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
##  IN NO EVENT SHALL THE AUTHORS AND COPYRIGHT HOLDERS AND THEIR
##  CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
##  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
##  LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
##  USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
##  ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
##  OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
##  OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
##  SUCH DAMAGE.
##

require 5.008;
use strict;

use Data::Dumper;
use CGI;
use CGI::Cookie;
use CGI::Session;
use CGI::GuruMeditation;
use DBI;
use DBD::Pg;
use MIME::Base64;
use XML::Simple;

#   configure optional debugging
$Data::Dumper::Purity = 1;
$Data::Dumper::Indent = 1;
$Data::Dumper::Terse  = 1;

#   program name, version and date
my $progname="registry-ui.pl";
my $progvers="0.2.1";
my $progdate="28-Nov-2005";

#   determine path to OpenPKG instance
my $PREFIX='@l_prefix@';
$PREFIX=$ENV{OPENPKG_PREFIX} if ($ENV{OPENPKG_PREFIX} ne "");

#   configuration
#
my $cfg = {};

#   user identification
#
#   constant  - no authentication at all, assume single constant default user
#   naive     - no authentication at all, ask for username and just trust it
#   basicauth - believe in webserver's basic authentication
#   ase       - OSSP affiliation services environment
#
$cfg->{identification}->{mode} = "ase";
$cfg->{identification}->{default} = "anonymous";
$cfg->{db}->{registry}->{username}="registry";
$cfg->{db}->{registry}->{password}="registry";
$cfg->{db}->{registry}->{tablespace}="registry";
$cfg->{db}->{registry}->{host}="127.0.0.1";
$cfg->{db}->{session}->{dbfile}="$PREFIX/var/openpkg-registry/ui/session.db";
$cfg->{canvas}->{url} = "http://meta.openpkg.org/?path=project.registry";
$cfg->{canvas}->{urlcache} = 1;
$cfg->{canvas}->{mark_head}="<!-- CANVAS: HEAD -->";
$cfg->{canvas}->{mark_body}="<!-- CANVAS: BODY -->";
$cfg->{page}->{default} = undef;
$cfg->{status}->{showuser} = 1;
$cfg->{status}->{showversion} = 1;
$cfg->{status}->{showsid} = 1;

my $ase;
$ase = undef;
if ($cfg->{identification}->{mode} eq "ase") {
    #use OSSP::ase::client;                                 # FIXME production
    BEGIN { require "./ase.pm"; import OSSP::ase::client; } # FIXME development
}

#   create objects
my $cgi  = new CGI;
my $myurl = $cgi->url(-relative => 1) || ".";
my $requestedpage = $cgi->url_param("page");
$requestedpage = $cfg->{page}->{default} if (not defined $requestedpage);
$cgi->delete(-name=>'page');
my $out = undef;

#   database handle and scratch variables
#
my $sql; # scratch SQL string database operations
my $rv;  # scratch return value for database operations
my $rc;  # scratch return code for database operations
my $sth; # scratch statement handle for database operations
my $msg; # scratch variable for message fiddling
my $dbh; # database handle
my $dbs; # database handle for session

sub dbopen($) {
    my ($db) = @_;
    my $dbi;
    $dbi = undef;

    if    ($db eq "registry") {
        $dbi = DBI->connect (
            "DBI:Pg:dbname=".$cfg->{db}->{registry}->{tablespace}.";host=".$cfg->{db}->{registry}->{host},
            $cfg->{db}->{$db}->{username},
            $cfg->{db}->{$db}->{password},
            {
                PrintError => 0,
                RaiseError => 0,
                AutoCommit => 0
            }
        );
    }
    elsif ($db eq "session") {
        $dbi = DBI->connect (
            "DBI:SQLite:dbname=".$cfg->{db}->{session}->{dbfile}, "", "",
            {
                PrintError => 0,
                RaiseError => 0,
                AutoCommit => 1
            }
        );
    }
    else {
        die "dbopen does not know how to handle db=\"".$db."\"";
    }
    return $dbi;
}

#   first check for pages which do not require database access
#
if    ($requestedpage eq "css") {
    $out = '';
    $out .= &viewhttp('text/css', '+3600s');
    $out .= &viewcss();
    print STDOUT $out;
}
elsif ($requestedpage eq "jpg") {
    $out = '';
    $out .= &viewhttp('image/jpg', '+3600s');
    $out .= &viewjpg($cgi->param("name"));
    print STDOUT $out;
}
elsif ($requestedpage eq "gif") {
    $out = '';
    $out .= &viewhttp('image/gif', '+3600s');
    $out .= &viewgif($cgi->param("name"));
    print STDOUT $out;
}

#   pages below require database access
#
$dbh = &dbopen("registry");
if (not defined $dbh) {
    &printprettyerror("Registry database backend unavailable", &prettydbi());
    goto CUS;
}
$dbs = &dbopen("session");
if (not defined $dbs) {
    &printprettyerror("Session database not accessible", &prettydbi());
    goto CUS;
}

#   establish CGI Session object
my $session;
$session = undef;
CGI::Session->name("registry-sid");
$session = new CGI::Session(
    "driver:sqlite;serializer:Storable;id:uuid", $cgi, { Handle => $dbs, TableName => 'session' }
);
if (not defined $session) {
    &printprettyerror("Session handling failed", "");
    goto CUS;
}
$session->expire("+3600s");

if ($cfg->{identification}->{mode} eq "ase") {
    #   establish ASE object
    #
    $ase = new OSSP::ase::client(
        -server  => "http://registry.openpkg.org/ase",
        -cgi     => $cgi,
        -mode    => "page",
        -session => $session,
    );
    if (not defined $ase) {
        &printprettyerror("Affiliation Services Environment", "");
        goto CUS;
    }

    #   give it a chance to run
    #
    if ($ase->responsible()) {
        if($ase->action()) {
            print $ase->response();
        }
        else {
            &printprettyerror("Affiliation Services Environment", $ase->error());
        }
        goto CUS;
    }
}

#   continue to the pages that require database access
#
if    ($requestedpage eq "login") {
    $out = '';
    $out .= &viewhttplogin("association");
    $out .= &viewhtmlhead(-menu);
    $out .= &viewlogin();
    $out .= &viewhtmltail();
    print STDOUT &canvas($out);
}
elsif ($requestedpage eq "logout") {
    $out = '';
    $out .= &viewlogout();
    print STDOUT &canvas($out);
}
elsif ($requestedpage eq "asecomeback") {
    $out = '';
    $out .= &viewasecomeback();
    print STDOUT &canvas($out);
}
elsif ($requestedpage eq "association") {
    $out = '';
    $out .= viewhttpauthrequired("association");
    $out .= &viewhtmlhead(-menu);
    $out .= &viewassociation();
    $out .= &viewhtmltail();
    print STDOUT &canvas($out);
}
elsif ($requestedpage eq "dropxml" and not &uao()) {
    $out = '';
    $out .= &viewhttp();
    $out .= &viewhtmlhead(-menu);
    $out .= &viewdropxml();
    $out .= &viewhtmltail();
    print STDOUT &canvas($out);
}
elsif ($requestedpage eq "dropxml" and &uao()) {
    $out = '';
    $out .= &viewhttp("plain/text");
    $out .= &viewdropxml();
    print STDOUT $out;
}
elsif ($requestedpage eq "ase") {
    $out = '';
    $out .= &viewhttp();
    $out .= &viewhtmlhead(-menu);
    $out .= &viewhtmltail();
    print STDOUT &canvas($out);
}
else {
    if (not defined $out) {
        $out = '';
        $cgi->delete_all();
        $out .= &viewhttp();
        $out .= &viewhtmlhead(-menu);
        $out .= &viewhtmltail();
        print STDOUT &canvas($out);
    }
}

#   die gracefully ;-)
#
CUS:
undef $ase;
undef $session;
undef $cgi;
undef $dbh;
exit(0);

#   check whether user agent is openpkg-register
#
sub uao()
{
    return $cgi->user_agent() =~ m:^openpkg-regist(er|ry)/[01][\.ab]\d+[\.ab]\d+$:;
}

sub httpheader ($$$)
{
    my ($type, $expires, $refresh) = @_;
    my $header = {};
    $header->{type} =    $type    || 'text/html';
    $header->{expires} = $expires || '+1s';
    if ($refresh) {
        $header->{refresh} = "$refresh; $myurl";
    }

    #   determine HTTP response cookies for session persistency
    if (defined $session and $session->is_new()) {
        $header->{cookie} = $cgi->cookie(
            -name    => $session->name(),
            -value   => $session->id(),
            -expires => sprintf("+%ds", $session->expires()),
            -path    => $cgi->url(-absolute => 1)
        )
    }

    return $header;
}

sub viewhttp ($$$)
{
    my ($type, $expires, $refresh) = @_;
    my $header = {};

    $header = &httpheader($type, $expires, $refresh);

    return $cgi->header($header);
}

sub viewhttpauthrequired($$$$)
{
    my ($page, $type, $expires, $refresh) = @_;
    my ($header, $username);
    
    $header = &httpheader($type, $expires, $refresh);

    $username = &identifyusername();
    if (not defined $username or $username eq '') {
        if    ($cfg->{identification}->{mode} eq "ase") {
            $header->{redirect} = $cgi->redirect(-uri => $ase->url(-action => "login", -mode_during => "ase", -mode_after => "association"));
        }
    }
    return $cgi->header($header);
}

sub viewhttplogin($)
{
    my ($page) = @_;
    my ($header, $username);
    
    $header = &httpheader();

    $username = $cgi->param("username");
    $header->{cookie} = $cgi->cookie(
        -name      => 'username',
        -value     => $username,
        -expires   => '+3600s',
        -path      => '/'
    );
    $header->{redirect} = $cgi->redirect("?page=$page");
    return $cgi->header($header);
}

sub viewhtmlhead (;$)
{
    my ($menu) = @_;
    my $head;
    $head = "";
    $head .= "<html>\n";
    $head .= "  <head>\n";
    $head .= "    <link rel=\"stylesheet\" type=\"text/css\" href=\"$myurl?page=css\"/>\n";
    $head .= &printjscheckallboxes();
    $head .= "  </head>\n";
    $head .= "  <body class=\"registry\"><div class=\"registry\">\n";
    $head .= "      <img id=\"logo\" src=\"http://www.openpkg.org/openpkg.gif\" alt=\"\" width=\"148\" height=\"80\"/>\n";
    $head .= "      <h1>OpenPKG Registry</h1>\n";
    if ($menu =~ m/-menu/) {
        my $td;
        $head .= "      <table class=\"menu\">\n";
        $head .= "          <tr>\n";
        $td = 0;
        $head .= "              " . &viewdropxmlform() . "\n";     $td++;
        $head .= "              " . &viewloginform() . "\n";       $td++;
        $head .= "              " . &viewassociationform() . "\n"; $td++;
        $head .= "              " . &viewlogoutform() . "\n";      $td++;
        $head .= "          </tr>\n";
        $head .= "          <tr>\n";
        $head .= "              <td colspan=\"" . $td . "\">\n";
        $head .= &prettyauthinfo;
        $head .= "              </td>\n";
        $head .= "          </tr>\n";
        $head .= "      </table>\n";
    }
    if ($cfg->{identification}->{mode} eq "ase") {
        $head .= $ase->canvas(-part => "head") . "\n";
    }
    return $head;
}

sub prettyauthinfo ()
{
    my ($html, $username);
    $html = "";
    $username = &identifyusername();
    if ($username ne "") {
        $html .= "authenticated as <b>" . $username . "</b> ";
    }
    else {
        $html .= "you are <b>not authenticated</b>\n";
    }
    if    ($cfg->{identification}->{mode} eq "ase") {
        $html .= " via ase login";
    }
    elsif ($cfg->{identification}->{mode} eq "basicauth") {
        $html .= " via basic authentication";
    }
    elsif ($cfg->{identification}->{mode} eq "naive") {
        $html .= " via naive input";
    }
    elsif ($cfg->{identification}->{mode} eq "constant") {
        $html .= " via constant setting";
    }
    else {
        $html .= " through magic";
    }
    return $html;
}

sub viewhtmltail ()
{
    my ($msg, $html);

    $html = "";

    if ($cfg->{identification}->{mode} eq "ase") {
        $html .= $ase->canvas(-part => "body") . "\n";
    }

    $msg = "";
    if ($cfg->{status}->{showversion}) {
        $msg .= ($msg ? " &nbsp;|&nbsp; " : "" );
        $msg .= $progname . "&nbsp;" . $progvers . "&nbsp;(" . $progdate . ")";
    }
    if ($cfg->{status}->{showuser}) {
        $msg .= ($msg ? " &nbsp;|&nbsp; " : "" );
        $msg .= &prettyauthinfo;
    }
    if ($cfg->{status}->{showsid}) {
        $msg .= ($msg ? " &nbsp;|&nbsp; " : "" );
        $msg .= "sid=" . $session->id;
    }
    if ($msg) {
        $html .= "<p/>\n";
        $html .= "<div class=\"status\">\n";
        $html .= $msg;
        $html .= "</div>\n";
    }

    $html .= "  </div></body>\n";
    $html .= "</html>\n";
    return $html;
}

sub viewcss () {
    my $css = '';

    $css .= "/*\n";
    $css .= "**  registry-ui.pl - OpenPKG registration user interface\n";
    $css .= "*/\n";
    $css .= "\n";
    $css .= "BODY.registry {\n";
    $css .= "    background:      #cccccc;\n";
    $css .= "}\n";
    $css .= "DIV.registry {\n";
    $css .= "    font-family:     sans-serif, helvetica, arial;\n";
    $css .= "}\n";
    $css .= "DIV.status {\n";
    $css .= "    font-family:     sans-serif, helvetica, arial;\n";
    $css .= "    font-size:       66%;\n";
    $css .= "}\n";
    $css .= "BODY.registry DIV.registry {\n";
    $css .= "    background-image: url($myurl?page=jpg&name=bg);\n";
    $css .= "    border:           1px solid #000000;\n";
    $css .= "    padding:          20px 20px 20px 20px;\n";
    $css .= "    font-family:      sans-serif, helvetica, arial;\n";
    $css .= "}\n";
    $css .= "img#logo {\n";
    $css .= "    display:none\n";
    $css .= "}\n";
    $css .= "BODY.registry img#logo {\n";
    $css .= "    display:block\n";
    $css .= "}\n";
    $css .= ".registry table {\n";
    $css .= "    padding:          0px 0px 0px 0px;\n";
    $css .= "    font-family:     tahoma, sans-serif, helvetica, arial;\n";
    $css .= "    font-size:       100%;\n";
    $css .= "}\n";
    $css .= ".registry tr {\n";
    $css .= "    padding:          0px 0px 0px 0px;\n";
    $css .= "}\n";
    $css .= ".registry td {\n";
    $css .= "    vertical-align:   top;\n";
    $css .= "    padding:          0px 0px 0px 0px;\n";
    $css .= "}\n";
    $css .= ".registry input {\n";
    $css .= "    width:            100%;\n";
    $css .= "}\n";
    $css .= ".registry H1 {\n";
    $css .= "    font-family:     tahoma, sans-serif, helvetica, arial;\n";
    $css .= "    font-weight:     bold;\n";
    $css .= "    font-size:       300%;\n";
    $css .= "}\n";
    $css .= ".registry H2 {\n";
    $css .= "    font-family:     sans-serif, helvetica, arial;\n";
    $css .= "    font-weight:     bold;\n";
    $css .= "    font-size:       140%;\n";
    $css .= "}\n";
    $css .= ".registry A {\n";
    $css .= "    text-decoration: none;\n";
    $css .= "    color:           #6666aa;\n";
    $css .= "}\n";
    $css .= ".registry INPUT {\n";
    $css .= "    background:      #ffffff;\n";
    $css .= "}\n";
    $css .= ".registry TD.browse {\n";
    $css .= "    width:           100%;\n";
    $css .= "    background:      #d0d0d0;\n";
    $css .= "}\n";
    $css .= ".registry TD.browse SPAN.title {\n";
    $css .= "    font-weight:     bold;\n";
    $css .= "    font-size:       200%;\n";
    $css .= "    color:           #000000;\n";
    $css .= "}\n";
    $css .= ".registry TD.query {\n";
    $css .= "    width:           100%;\n";
    $css .= "    background:      #d0d0d0;\n";
    $css .= "}\n";
    $css .= ".registry TD.query SPAN.title {\n";
    $css .= "    font-weight:     bold;\n";
    $css .= "    font-size:       200%;\n";
    $css .= "    color:           #000000;\n";
    $css .= "}\n";
    $css .= ".registry TD.view {\n";
    $css .= "    width:           100%;\n";
    $css .= "    background:      #d0d0d0;\n";
    $css .= "}\n";
    $css .= ".registry TD.view SPAN.title {\n";
    $css .= "    font-weight:     bold;\n";
    $css .= "    font-size:       200%;\n";
    $css .= "    color:           #000000;\n";
    $css .= "}\n";
    $css .= ".registry TD.result {\n";
    $css .= "    width:           100%;\n";
    $css .= "    background:      #d0d0d0;\n";
    $css .= "}\n";
    $css .= ".registry TD.result SPAN.title {\n";
    $css .= "    font-weight:     bold;\n";
    $css .= "    font-size:       200%;\n";
    $css .= "    color:           #000000;\n";
    $css .= "}\n";

    return $css;
}

sub viewjpg () {
    my ($name) = @_;
    $name .= ".jpg";

    my $jpg = {
        "bg.jpg" => <<'EOT'
/9j/4AAQSkZJRgABAQEASABIAAD//gAqQ3JlYXRlZCB3aXRoIEdJTVAgYnkgUmFsZiBTLiBFbmdl
bHNjaGFsbP/bAEMABAIDAwMCBAMDAwQEBAQFCQYFBQUFCwgIBgkNCw0NDQsMDA4QFBEODxMPDAwS
GBITFRYXFxcOERkbGRYaFBYXFv/bAEMBBAQEBQUFCgYGChYPDA8WFhYWFhYWFhYWFhYWFhYWFhYW
FhYWFhYWFhYWFhYWFhYWFhYWFhYWFhYWFhYWFhYWFv/AABEIB9AFAAMBIgACEQEDEQH/xAAbAAEB
AQEBAQEBAAAAAAAAAAAAAgEDBAUGCP/EACUQAQEBAAIDAQEBAQEAAwEBAAABAgMRBCExEkFRYSIT
FDIFcf/EABcBAQEBAQAAAAAAAAAAAAAAAAABAgP/xAAaEQEBAQEBAQEAAAAAAAAAAAAAARExIUES
/9oADAMBAAIRAxEAPwD+/gAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAZdSfaDRP/wAmP9JyYv8ATRQy
alaAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA
AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAM7jQAAAAAAAA
AAAAAAAL6BO71Hk8rm6z6rv5W5MenzObdurGbVkTyc+/19bw+Rrv25anaL3PjKvq+Nzy/a9WeTN/
r4OOTcerxue/2rKY+tLK15uHnzf6753NfGtZUAoAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA
AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA
AAAAAAAAAAAAAAACdbkc98+ZKmi9ckiNeRif14vJ8iW+q8nJyat+pq4+r/8AYzdfXXHJLHxMb1+u
+3s8fm667qauPpwceLmzZ06y9xtloAAAAAAAAAAACeW9ZVb1HDyeSfjpKPL5PJ36eTU70vktuzr0
w259MuXXqssBxuWX9T463LPyCePk3P693ic8n/6rw6z/AIm3U+UlR9zHNmumbL8fH8fmueu693j+
Rn+tSpj1icbmvimkAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA
AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAABmr0575pkF6105
7585efyPJjxc/Lb/AFm1cezm8me3j5+W6vquNtv9IysZr9X+tk/1UjeoLif43Nvf069tQrtwcll9
17+HyJZ0+X7Vxaub9alTH2c7lU8Pj8/fp68b7jUqWLAVAAAAAAAC+oCeW9YfP8jk/wDVj1+Ryf8A
ix83l7vJ2xa1GX/9dqnxk+NRRshPjQT+T8rkb1AcrlGsu9ibkR59Sz43GtS/XTWUayK9vi+RMz3X
qxzzT4vWpfr0cHL+ftXUx9fN7jXk4PJnXXb1Y1+p21Ky0BQAAAAAAAAAAAAAAAAAAAAAAAAAAAAA
AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA
AAAAAAAAABm9dQGp5Nflx5PI/P8AXl8jy+59TVx35vIkeTm5+/648nJ+/wCufXbOq3k1df1HVdJl
v5RUZipFflsgIF9HQJ6rZFflsgJ6LmrgCeK3N7ezx+efHk1O4zjv5vfYj6+N9xbw+Nz9+u3rxvuN
ypYsBUAAAAE8nrFVfUcOfl/82JR5PI5P/XXbjZ3W8vvfbZ8YaTI3poANkOoDZ8AAK2RvQOdibl1s
TYDjrKNZvb0XKdZBz493Ovr3eP5UmZO3h3hme836Sj7PHyfqOk+Pl8Hkdeu3s4ef9f1qVMegZL3G
tIAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA
AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA48vL+f683N5Xrrt5+Xn/V+uOr3WNaxfLzfr+uN
naphWcoqMYdJlUyqQEzLev8AigEijqiJ6b0rqsBnTFJFAb0Ixmp3FdRgJ47+K9fi8/bybhx6/FIP
r412t4vE5v117ezN7biVoCoAX4CeT/8AFfO8nk/99PZz8nWbHzue98jNqxs93tUicfFZ+Mq0ABsZ
02egLCRoAAADZAZ1/wAZYss/4DjrLnrD0WJsB5vz1XXi5fzW6y5ax1QfQ8fyf167evN7j4vHr8ae
vh8rv121Klj6A48XL+nZpAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA
AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAHwvz3WzKpFSObaZFSNkbI
BI3qtkBGdNkG9AwABmml+AlNUdCpUdAgy/GstBOkaVakV08ff4r6Hi8n66fKt6d/G5fz0sqV9Ycf
H5P07NsjNf8A5rUcuv8AzQeLyt9a6ebV7qvL1/7c5XOtLzVyuefq4CmyMbkGgAAQBsjYAAA3LWZa
DLE2LZYDnYnWXWxlgPPrCZPz7d7E6z2Bxc9zXs8bm/f9fPuelcXJ+Kspj7HY8PD5Fv16uPcv9a1M
dAFQAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA
AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAB8eRUjZGubbJGioIzpvUAAAC/Etv1NoNZaS1oM6p1Wg
JsK3XxOgYALidRK9IoMs7JrpgD2+FyPdx67fGxyXNfR8Hk7nurKleu/Hk8nfXcerVn5r5nm760tS
OPPe9IlNXtjLS8326ZrlPi80HSVqJVSiLhGT40G2EJWgAAAANlOmX6ChkrQGWNbZ6BFibHRnQOWs
o1l3sTcg4TVy7cPPU6yi/wDkH0fH5P1/XeWV8nj5bl7fF5f19rUqWPSMmpf61pAAAAAAAAAAAAAA
AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA
AAAAAAAAAAAAAAHywkbI5tEjQAAAZad/4dUGDejoGAAAAzSdfVX4nQIv0ZazsVuk6bakAZ3/AKdh
hfrt4/LcuB+ugfT4ubvF7rw+brvac8tkTu/q9moZalsoqsqlQ2UHSVcrlKvNB0lUjK58EGz4yKAG
xvUBJ0oABvQJ69tKKNnxrI1BnXssaKJZYsBy1lGsu9iNRB59Z6M8lz8XuOeoD08HNb9r18XJLPr5
X6srrxcupYso+rL2OHj8sufddpqX+tay0BQAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA
AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAB80BzaBnZ2DWWl7JOg
JGgAX4AJnQABay1NoKt9OeqXSKKWp7b1ey9IMO4ys7ijbU2lqbQbaztN0nv/AARute18d9ON7tXi
UHXtsrn7bKK6StRKqVBUq81zlbKo7ZrplxzXTNEdMtTKqAqfAgAClGSNBBlnbZBlBontXZgALAb0
xUqianUVfrNJRz1HPUdtIsQcdZTe47WI1AZjl1K9XBzf7Xj1GZ1qaJR9fj3LHSPncHL19r2cXLmz
pqVLHUJe4NIAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA
AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA+Z3C1g5tDZCRoAAAADNNTQGWstTqgaqLfbakUGWnYNT
qmqjVAtTay6RrQLuka2jWkW0xFavbcIk7dMQGz67cc9OUnVdeP4DbE2V0n1vQrnG9lynqygtsqJf
apQdM10zXHNXmg7SqzXKVUojtmqcs1coLljLWdt7gNlalsqhe+zstYBQbPgNnwBAAXQTS1mvhRjN
NLO0EaTfi7E34CNRGovTL8FRLZXXx+Wzf1yreLN/foR9bh3LiOjy+Nm/mPTn43ErQFQAAAAAAAAA
AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA
AAAAAAAAAAB8tsjRzaAAAAC3pnbLQLU6oygy/E1SaKWptKm/UCstZquetqL1pz1tO+Ry1rsRe9ud
0m0kUaSNkVIBmOmficxciDL9deO+nLX1eKDpF5cpXTNFV1E6z6UCONz1Tvqul9uesiqzXTNcM3qu
mNA7ZVlzzpcoLzVyucVKIvtqJVSg1SWyg0ABU+Jblfg0ACptbr4yoDL8ZQAbG2KIsTr6tGkHPSdL
1/XP7roDOe9PV43DfvSfH4u7K9/DmTPxZC1vHmTKgbZAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA
AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAfNGW+2ObShL
e6Dam0tTaDbWMtZaK21lqbploKtTpnbLeoBanWuojfJ05b5BG8nI4637ZvXdTPdUb32ddtzlcyDn
MqkX0SIMkVIKgGYuRmYoHLf1WfjN/W5+A2fXTNc59XkVcre/SFSoDL8aAi5ZJ06M1O1DOl505XPT
c6B6M1UccadM0FyqlQSiOsrUStlBcrUgKJ9Z21RXcZ3GCDbWX4y0l7BjemgAAMv1z06ac9g53/F8
XF3qVmc96e3x+P18WCuDHWXafGZnUa3GQAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA
AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAHywZa5tNt6Z2ztnYrWX6
y1OtCN1U2ptYKrtnbLXPe+gXrXTjycqN8jlrXYit77qLezv2Azr2qQkVn6o3MVGT61BtrGydt/IJ
Vn22Zb10DZ6alWVHLk+r4/jOSe24QFRKp8Fimz6yUiCopMqgAD08ZqdouenQsBEvS86Rcs+KO80q
V55p0xoR2lVK5SrlBcv+NlRKqUFEZK2AoAEk+gChkrQBufhpROq47v8A6Vya9mM/qyoO/jcfcler
jz+Yjxp1l1bkSgCoAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA
AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA+T2y1Np3XNprLWWo1RVXSbe09stBSN66TvfT
jvkEdN8nTlrfaNa7Z32DdXtg2RRkipPTei+oDFRKp8Bs9Tts9mYqZAjbf+LzlX4MEZXM9tmFSdKO
dyzqu3XZ+Ewebf1eJ6dNcftsx1DBx19ack9iASjZBW5rpPjnPSpQUDehGDeixFYz89t6qpActTol
6ddZ7RrKjc6dM15++l50DvKqOWdKmgdJVSucrZRHWVtc5W9g0Z2Sg1srAFSp1ocuXXsDfuvV4vH3
HDhz+nu8fPUWFXjPUaDbIAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA
AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAD4tqey1Nrm221Oqy6c+TYN1rpz3yOe
9pttEbrfaLezTFgNyyKKNy1ICgbkG5jpnKcyKBecxUkRmrz3VFZisxmZXTGaDJD812zlv4gI489r
/Mb118bActZTY62M6B595c7Hq1lOsTpMHn6pIvU9kyzi6yRWW9EUblUiY0G9QsI0E1uW34T4AnUd
JGWA46yi+v476yjWQRNLzpFnTJbKDvNKlcc1cojrK2X/AK5TTf0DrK2OUq80HQTK3sDV6cOX6vk0
nM/VB6/Bz3l68zpw8OdR6G4lAFQAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA
AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAB8C1G9SJ3tx3uubat7c9atZb
WCCb2oNGSN6gAykrL9FFBPgg2VsqW5ntR0yvMTh1x0DcZduPDOPLvxxRuOOOn4jcR06Ecvyyx2sT
chrnI3pXTKDOjqtkrpnIOX5TrPp6LlOshrx6x7Z+f+PVcJ1mdJivP1/wmfbrc+24wg53NTZXpuPS
fwYOMla7fhO8g534ZjelZAkbY2dMv0GWJsWywVy1lz1l6LEayDh7lVmq1lFlgOkvolc5fa+wXKqV
zlbKDrKXSJf8LRE8uvbv4mf082/de/8A/nZ9RYPTwzqOhJ0NsgAAAAAAAAAAAAAAAAAAAAAAAAAA
AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA
APyetVIObQAQCgtGTtoAykjVZAzFTLc57dcYUcvxT8V6Jlv5/nQPPJV8Xf6dP/itVjjsoOvHPTtx
xHHOnbHQOmYuRGXTMVlljOnTosBy1GdOljOkEyOmZEydrzFDplyvo6By1n056j0anpzufaK4zPtc
zFfksFTOjqEntUgJ1HPcdtRz1lBxsJ9dLln5qCRtlZ1QC/G9VlBI2wk9gy59Oe8OzLOwebWUzuPR
vLlvIpK2X/HP3K3NB1lLUys3QXmdvoeDOpHh4J2+h4k6ixK9ADbIAAAAAAAAAAAAAAAAAAAAAAAA
AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA
AAAD8iA51oAUGWtrOgJ8ayTpsBsjZm9tyrPQOvFPTrnpyw7Yzao2TtWMXtucO2JAM5nXxn59/FyN
6ET0rEbMrxkNbxx2zE5yvMVG9MsaSdqJsTculOogjMXINnxQ6jegBlibFsvxBzsT06aQis6jWyNo
Js7ZcrkLA1ysT1HXURqCueoTK7CQEflFjtYjUTBz6h0rpqCOjqKsYCbEby6pv0Hn3hH5seqyVOsd
iuGfTN1e8Vx3LAezwvfT6PDZHyvE5Jl7MeRIsqPdKPJnys/6ueTGtTHoHD/7EP8A7ENMdxwnkRs5
pTUx2ETklb+zRQyVqgAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA
AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAD8lYx23jpzsYaTJ23psis+gT+Wfl0thIDn02el9
NzjsEyNmfbtx8TpngUTw47eriz1E4x+XbEAmWzPtecqmfYjM59NmF5i+vSo55yvOWyKkAkbBU+KM
6bAATfqk36g3LWZ+NtUBnZ2DS/GdloJvxkntTZEGdM6WAkUywE6c9R018Z12ixzmW9L6ARYjUdbE
WCudPyv8nSYOdjK6VNhgixPS01AZanWuka5Owbyajhufpd9tzgVzxiq/OnWRoOXWv9rZ3/rom/QZ
/wCv9b7/ANARvd/1Wd2X6hlgPTnn6jc+RI8fV/1l7NH0M+TFzyY+ZO5/VTV/1f0Y+pjnmnXOu3yu
Ll/P9d8+VJ/VlTHvHinmT/Xbi55r+rqY7jM3tqgAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA
AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAD4HJx9uPJx9PbrLnyY7c2nhs
6rJXo5ONxuegMzt0zhPF9enix20J4+H9O2fG6/jtxZ/Ltn2Jrhjh6/jpnHX8dTrsNc//AI+1Zx06
SN6VE5ipGydNkAkVIRUgMFHUUS2VvUZYyNZawAGydsqyCp8ZZ2S9Hah0dKlamiOmL0i/VGyNZKqQ
GDemyAlnbdf1KAQALGdNb0ip6TZ7WzoNT16Tp0sc9+oKnSNU3vpx5NpaK3rpy3yJ1vtH1ButfomW
5wuQVmc9LkJGyCM6h0rqHQJ6ZZ/xfTAR0dKsYKyyMX0ywE2MsbQNR0dLsZYYIsTYvSLQZ1f9d+Dl
/H9ee6ZNA+jjy+v69fj8v/yR8zx+H99Po+Lx/iNRmu4DSAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA
AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAPmXKNZdqmxzaefeHHk4
3s1lz3kHi6/K+Pm/LrycfbhvHRo9nj8n7erD5OOS4e7xOX9ddtD1RUTiqVlQT2qQGSNCfVFT42fG
HYKGStAZr40BMnbemkAOuytl9AnplVampoqfBnbCje4wD0bIdnbFFSqnxGfq58BOkq2yRAka3o6B
nX/BUiSDKm3qK1fTzc3L0lWK3ydOPJyuXJyduWt91NVe+T2jWu0z2vOUVMz7XnPSs5XMiJkVI3ro
BnTRvQMFSHQJZVWMoJAAZfjU2gzTDVTaKq6/6nVZanVA1Uap3a7+Nxf/ACfwHnk7ejxuD9fx68+H
P8d+HhmP4uJqfG4fxHok6JOmtsgAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA
AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAPngObSbE2OjLAcdZc947ejURcg8nJxpzyXje
reXHk4wenxOb9fa9eLL/AF8mauL6eng561KmPo5XHHg5JZ7rtL38VAV1GdeyjCRQB8Z22/EqN7O2
CaN7bKkUXbOmVnbEC0GyGBIyr69Jv1RgDIAAKlS3KwbtmfqrCRRoWsuoBb6cuTfX9RzcsnyvLy81
ZtXHXl5uv68vLyXVTvdtZJ2zqlvbZlsyuQE5y6ZjZFSAyRUAGU6rW9AmTpUh02A2RlimUE2Jv1V+
I0CazuGk2g21Npam0UtTaaqPdoN1azHer8rt4/HdX3Hu4/FzJFkS15fG4P19j3cHjzHS+PimXRqR
LSAKgAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA
AAAAAAAAAAAAAAAAAAAAD542wkc2iQsaAmxNjoywHLWXPeXfURqA828Jvefj0ay56yK548jedyPp
+HyzWPdfL3j+meXeNdT4sqWPuzWb8rXzPG8i3ruvfxcubn77aZdAnv4WKJtCfVJRPVb1WyN6QT1S
RtIoysboyoSKibTugq1N+s7ogqRvRK0wZ0dNDBOmZbpPcn1R0ZdSI1yZn9ebyeeT5U0ejfJn/Xm5
+br5Xl3z6qP3rX1nWsXy8tqLbW5na5lBOc9rzlsipAZIqRsjQZI0ACCp8BkjWdxsvYBPoLopmmJt
QNVz1W6056oM1U2mqi0VVqbb/Eyat+O/DxW33AcsZ1dfHq4ODu+478PBOvj0Y45Isiani4M5nbrJ
02DbIAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA
AAAAAAAAAAAAAAAAAAAAAAAADwsn2sHNpQyVoAAMv1FjpU0HOxNjpqJsBy1ly3iPTY57gPNdazfT
r4/PqbndZvMctSy+l0fb8bnzcfXbvv2+D4vNvPJ7vp9fx+fNxJ37alSx2n1U+JzqVU+FRrLS1lpI
FrO6xvRRgqM0QZG9Ny3owT1WL6LDBEq5U2E7UWy3pl1JHLk5s/6DprWZ9cPJ5syeq4+Tzzr1Xh5O
XVv1m1cdebn1/K4Xk3q+0zurxllTPddePJnLpnIpnK5CRUgjAACfRsgNAAbawAC1nYK7Zam1F0C7
pN0i6RrYL1pz3pOraz8at9CsvdvpfHx7tdeDgt+x7+Dhz+fcWRNeXg4P9j28XDmRecZio1ImsmZG
gqAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA
AAAAAAAAAAAAAAAAAAAAAAAAPmtjRzaGysAUMlb3ACwATU2OlTQRY56jtY56gOOojeXXURYK42de
4vg5d517pqI1P8JR9Tx+eWSdvXjcs9PhcO9Z19e/xvI+dtSs499qU45ZqLkaQkV0SNTRnXpOl34i
wgZqpUKlUUM79M1voG66kct8uYjn58zPTwc/Nbr1WbVx6vI5516rwcvLq36ne9WpktrOq261f6Zz
e1Zy6ZyKnOV5yvOVyCJzlchIA2RpPgDOmyN6YAAAAAzstTbAbanVZrSNaBuqi6ZdJvv4K26Tc618
Vji1p6uDgs+wR5+Lh1Xt8fgknuO/DxST3HWZkakTUcfHmT46SSA0gAAAAAAAAAAAAAAAAAAAAAAA
AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA
AAD53bUStlc2lDO2gAA3LWZ+tAZr61l+gm1Om36zXwEajnqOt+J1AcdfUV01E2IqLDGrm/W2MsUe
rxvI/P2vdw+Tnfx8TUv8rr4fJeO+61KmPuzXang4/Mn+vRxc80rLuyxmb2oEWHTbf68/P5Mx6Udd
8kzHl5/InXXbhz+T38rycu7q/WLVx05uW3X1y92smba6Yyiszl0zlWcrzkGYy6ZyZipAGwsJQazR
awFT42JipQalumABU2gruMtTanWgbrXpz1pm9OW99A6XTnvkjlvlcrvsHo/X6+O/j8V1Xl8b7H1f
Ak9EHTx+L8/Y9GcyfxUkG8ZAFAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA
AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAHx5pU04Z0vOnNt3lbK55qpR
HTsR22X/AKC5WolbL0CkttTaBr6m/GgJvxN+KsTYCNRNi6m/BUWJsWzQIsRvLppOqDnnvP8AXp8f
yfx/Xm3XLdXUfY4/On+r/wDvT/Xws6s/q88n/TTH2d+bOrO3z/L5v3r68uuS/wCp/fv6aO87v9Vn
Ljjk6dJzdIO+YvLzTm/6vPOD1ZXPry55/wDq884PS2VwzyrnJ2DpaxP6b+v/APAaJ77AU2Vk+MtB
XZantloNtTrTNac97BWtuetufJvpy3yA6cnJ048nJ2jWu0Wewbb3/TPojcg9Hi/Y+t4H8fJ8X7H1
vA/izp8e0BtkAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA
AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAB+czv26Z046z0TfTm09WdOkryce3bOgd5Wy
uWdLlB0lbK5yq7BaaztloKEdl0DbU6rLpGtA3VRqo1tF5BXTWunO8nSbrtN9iK1yIuzWUUG3SdNT
foM6LOmwvxRGmKABlrf6g2Nl6J8AVNNm+kHQOs5ulzn6/rzdRvQPXPI/6rPkf9eKfVSg9+OXv+um
d9vn55LF557Ae+WFryZ5+/66Z5OwdrUa0i6RvYqt7ceTkTvkcta7EbvaLWaYuDY0nxsgNnwn0kb/
AAHfxfr6vgT4+X4n19XwSdHsAbZAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA
AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAfn9Z7c946d7E6nbm088
vTpx7N4c76B6c6dM6ePG3bGwemVv6cZpv6FdLpl0i1lv/QXdI1ty3tF3QddcjneTtk9lgh32WHwt
BNAA18Rpd+JoJOu29NBmcs3F5TyAhN+rkZZ2olshIqQCRrY1BPUZYrqFgJkaRvUBnVbIS9tAuWdL
l7OoCZ6XN9MOoDf/AJGXXbOjoE1F+ulRr6omkjSQGyKkY2VBsCfW2A7+JPb6vhPleH9fV8P+LOle
sBtkAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA
AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAB8TU6RXXXSNRzbR9RvDofQebklynG69PJiVw5cdfBHT
j326TTyZtiv3Qd976Rd1zltXmdg2e2/lsnQB10lR0KkL0y0Q0m0tYDbWDegYyq6ZQblPIrKNgzPx
jctsBKp8Z0RRc+BPjZPSDBTKDBsjeoCRTLAYADZWsyvMBPVLHTqJ0Dnqpbr6wAO2dg1nZfjFF5rd
VGTVQerw77fW8N8fwb7j7HhfFnSvWA2yAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA
AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA+JaxErY5tqZ02f
VIIvaN5ldbEWA4bx0n8/8d7P9TcqmIzlU9FjBW9xqWz+hpay0L8ETayd9tbIDI3psKCRVTQE36pI
Nyja8p3ATn60AbIwn1QE+Nz8Y3INAAAgA3o6Bh02xnQNkXPjM/Wijnur1fTloRN+pUn+rAbI30zu
AX4wZ7BpZa3GdV24eDV/gK8DN7np9nw5ZHl8Hg667j6PHmRZEqwGkAAAAAAAAAAAAAAAAAAAAAAA
AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA
AAAAAAAfB1jUrJenv8ng9XqPFycWs3tzxozVyuM7i80VbNRs+AOeozpdSgmxOo6Ms9KOXVVPjbGA
dMrU36FAnRRA7KkG2sABKr8SDYnSonX0GAzsFSndZK3KjSAgpuWKnwAkG5+g0AUTZ/FFgJn1tp0w
Ga+ov1dRqiJrK21FqjWdsnuunHx2/ATJa68XDq/x28fxtd+4+j4vjyddwHk8Xxr37j6HB4+Z16ej
HHmT4qSRZE1OOPOfiwaQAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA
AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAABOsyz48/k8MufUeoslTB8bm8f
Utc5i5vt9nl45Z8ePyODv5GbGpXjlba3fFc/XO3pFbqsT32AovwgIkVYyzpFTfibF1KjOqzqqAT1
SxTNfAQNrBGWtgzsG3457q7XPkA7Tr6zutkUbn6vPxOVz4gNkJGgKZK0BuSRooAIAzsVqdVmqm0Q
1UVtrJO6CWzF1XXj4bqvVweJe5elHl4fH1a9/i+NZ13Hq8fx5J7j14xmT4uGuXDxZmZ6dpmRo0yA
AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA
AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAJ1mX+KAeTyuH9T08PN4+o+y48/H+v4zYsr4us/l
nb3eR41vfp5OTiuGWky/42VFvTZQWy/GSlAZ37amxFbfjD+F+CU7idVlrNVUZaSsZ2CrWEb0DEbX
Z0jkBP2tZn61RWV5RF5QaAAqMkVIDRUjNfUVkKM0oy1NpanVELU/Wydu3FwXVgOeOO6ejh8S29vV
4vi3/Hu4eKZnxZB5fG8br7Hu4sSZ+Kkk/jWpE06n+AKgAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA
AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA
AAAAAAAAAAACd57/AI8nk+P+vcj2lTF18Tn8e5/jz7/819vn4f28PkeL/wAZsV4ppXZycf5vSPiC
+6zus7p2K1mi1NojLUqvxICelAMnxfbJG9QGWue3Sxz1AQpljYoqfFZ+sn1WYg2KkJFfIKyTo7LW
AuVlTG2gWp1S1uM/oRFqscf6enh8b9/x6+HxOv4uDy+P4d1/Hu4PG/P8ejg4/wAuqyJqOLMzPiwa
QAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA
AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAc+Tj/AE6APDzeJ+vfTx8/B+f4
+1fjz83D+ozYuvh8k/N6Z29/k+N/eni5cXN+Mqi1nZZ7YDbWFnQAT6qQ66AAnsGa+oqtpBKsxK8q
NkVPSWyoLlGZV0DBt+JtBtvSfdJ7r0+Jwfu/Ac+Hj/f8e3x/F+eno4PF/P8AHp48flZE1z4eH8u0
nTRtAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA
AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAvwAc+Timo8vP4k6
7e5mp3OkxdfD8jh/N+PNqWX4+7z+PLO3h8nxuu7IzYr5/skXvFzr4i9/1BtrO2NnX+gdtlTW5Bm0
t0gBefiF5BvSpDM7igbmNqLWTu34DdWt4s3V+OvBxXWvj6HB4s6l6XB5/G8bv+PfwePMe18fFMus
WRLSeoA0gAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA
AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAABZ258n
FNOgDw+T4ueu4+d5HDZfj728zUefn4M2M2LK+BZqX4PoeX4/XyPFvj1P4yqFRPVn07Bm0Sdq1WZU
OlT6w7QXK3upxnV/j0cHDq6ncBz48auvj2+N4/f2PT43jZ/Pdj044pIuGufD42ZO3fOZIqTqDWMg
CgAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA
AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAyztoDnvi
zr68vl+Nnr1HuZrM19TF18HyOCz5Hm1mz7H3+fgzf4+f5fj331GcV83THXk4dT+JnFpBEldOPi1b
6jrxePq2en0vD8aSe4o8/i+Pb13H0eDx8zLpx8WZ/HSTpZE1OcyRQNIAAAAAAAAAAAAAAAAAAAAA
AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA
AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAyyVO+LN/iwHg8rx5b6jlnxv
fx9O5l+xn4z/AIzi64eNwZk9x3zmT42SRq4gAoAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA
AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA
AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA
AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA
AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA
AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA
AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA
AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA
AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA
AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA
AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA
AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA
AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA
AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA
AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA
AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA
AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA
AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA
AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA
AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA
AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA
AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA
AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA
AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA
AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA
AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA
AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA
AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA
AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA
AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA
AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA
AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA
AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA
AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA
AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA
AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA
AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA
AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA
AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA
AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA
AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA
AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA
AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA
AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA
AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA
AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA
AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA
AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA
AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA
AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA
AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA
AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA
AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA
AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA
AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA
AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA
AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA
AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA
AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA
AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA
AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA
AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA
AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA
AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA
AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA
AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA
AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA
AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA
AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA
AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA
AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA
AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA
AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA
AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA
AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA
AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA
AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA
AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA
AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA
AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA
AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA
AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA
AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA
AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA
AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA
AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA
AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA
AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA
AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA
AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA
AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA
AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA
AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA
AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA
AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA
AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA
AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA
AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA
AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA
AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA
AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA
AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA
AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA
AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA
AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA
AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA
AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA
AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA
AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA
AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA
AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA
AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA
AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA
AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA
AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA
AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA
AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA
AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA
AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA
AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA
AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA
AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA
AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA
AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA
AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA
AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA
AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA
AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA
AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA
AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA
AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA
AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA
AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA
AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA
AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA
AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA
AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA
AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA
AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA
AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA
AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA
AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA
AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA
AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA
AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA
AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA
AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA
AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA
AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA
AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA
AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA
AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA
AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA
AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA
AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA
AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA
AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA
AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA
AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA
AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA
AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA
AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA
AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA
AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA
AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA
AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA
AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA
AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA
AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA
AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA
AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA
AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA
AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA
AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA
AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA
AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA
AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA
AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA
AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA
AAAAAA//2Q==
EOT
    };

    return decode_base64($jpg->{$name});
}

sub viewgif () {
    my ($name) = @_;
    $name .= ".gif";

    my $gif = {
        "icon-dia.gif" => <<'EOT',
R0lGODlhCQAJAMIAAP////8AAP+ZAAAA/wCZAP///////////yH5BAEKAAcALAAAAAAJAAkAAAMS
eHqyLtAxGBW989KMbV+bxDgJADs=
EOT
        "icon-dot.gif" => <<'EOT',
R0lGODlhCQAJAMIAAP////8AAP+ZAAAA/wCZAP///////////yH5BAEKAAcALAAAAAAJAAkAAAMR
eLqs8+3JJaur8OCsLZtNmAAAOw==
EOT
        "icon-ok.gif" => <<'EOT',
R0lGODlhCQAJAMIAAP////8AAP+ZAAAA/wCZAP///////////yH5BAEKAAcALAAAAAAJAAkAAAMS
eLpKvMS9KGE8lTand7cVEyoJADs=
EOT
        "icon-x.gif" => <<'EOT'
R0lGODlhCQAJAMIAAP////8AAP+ZAAAA/wCZAP///////////yH5BAEKAAcALAAAAAAJAAkAAAMS
GHHM+g1KFd9qB2J7M50dh3kJADs=
EOT
    };

    return decode_base64($gif->{$name});
}

sub viewassociationform()
{
    my ($html);
    $html = '';
    $html .= '<td>';
    $html .= $cgi->start_form(-action => "$myurl?page=association");
    $html .= "<div>" . $cgi->submit('submit','association') . "</div>";
    $html .= $cgi->end_form;
    $html .= '</td>';
    return $html;
}

sub viewloginform()
{
    my $html;
    $html = '';
    $html .= '<td>';
    if    ($cfg->{identification}->{mode} eq "ase") {
        $html .= $cgi->start_form(-action => $ase->url(-action => "login", -mode_during => "ase", -mode_after => "asecomeback"));
        $html .= "<div>" . $cgi->submit('submit','login') . "</div>";
        $html .= $cgi->end_form;
    }
    elsif ($cfg->{identification}->{mode} eq "basicauth") {
        # nop
    }
    elsif ($cfg->{identification}->{mode} eq "naive") {
        $html .= $cgi->start_form(-action => "$myurl?page=login");
        $html .= "<div>" . $cgi->submit('submit','login') . "</div>";
        $html .= $cgi->textfield(
            -name      => 'username',
            -override  => 1,
            -size      => 50,
            -maxlength => 50,
            -default   => $cfg->{identification}->{default},
        ) . "\n";
        $html .= $cgi->end_form;
    }
    elsif ($cfg->{identification}->{mode} eq "constant") {
        # nop
    }
    else {
        # nop, misconfigured
    }
    $html .= '</td>';
    return $html;
}

sub viewasecomeback()
{
    my ($redirect, $username);
    my $html;
    my $header = {};

    $html = '';
    $username = &identifyusername();
    if (defined $username) {

        #   updating heartbeat or creating username
        #
        $dbh->{AutoCommit} = 1;
        $sql = sprintf("UPDATE reg_user SET heartbeat = now() WHERE ( username = '%s' );", $username);
        $rv = $dbh->do($sql);
        if (not defined $rv) {
            $html .= &printprettyerror("updating user $username", prettydbi());
        }
        elsif ($rv != 1) {
            $sql = sprintf("INSERT INTO reg_user (username) VALUES ('%s');", $username);
            $rv = $dbh->do($sql);
            if (not defined $rv) {
                $html .= &printprettyerror("inserting user $username", prettydbi());
            }
            elsif ($rv != 1) {
                $html .= &printprettyerror("creating user $username", prettydbi());
            }
        }
    }

    return $html if ($html);
    $header->{redirect} = $cgi->redirect(-uri => "$myurl?page=association");
    return $cgi->header($header);
}

sub printprettyerror($$)
{
    my ($marketingmessage, $technicaldetail) = @_;
    my $out;
    $out = '';
    $out .= &viewhttp();
    $out .= &viewhtmlhead();
    $out .= &viewprettyerror($marketingmessage, $technicaldetail);
    $out .= &viewhtmltail();
    print STDOUT &canvas($out);
}

sub viewprettyerror($$)
{
    my ($marketingmessage, $technicaldetail) = @_;
    my $html;

    $html = '';
    $html .= "<h2>Sorry</h2>\n";
    $html .= "<img src=\"?page=gif;name=icon-x\">&nbsp;an internal <b>ERROR</b> occurred and prevents further processing.<br/>\n";
    $html .= sprintf("<h2>Problem scope</h2>\n%s<br/>\n", $marketingmessage) if (defined $marketingmessage and $marketingmessage ne "");
    $html .= sprintf("<h2>Technical details</h2>\n%s<br/>\n", $technicaldetail) if (defined $technicaldetail and $technicaldetail ne "");
    $html .= "<h2>Please come back later and try again</h2>\nSorry for the inconvenience\n";
    return $html
}

sub prettydbi()
{
    my $msg;
    $msg = $DBI::errstr;
    $msg =~ s/[ ]*ERROR:?[ ]*//;
    $msg =~ s/ for user.*$//;
    return $msg;
}

sub viewlogin()
{
    my ($html, $username);

    $html = '';
    $username = &identifyusername();
    $html .= "Welcome, <b>$username</b><br>\n";
    return $html;
}

sub viewlogoutform()
{
    my $html;
    $html = '';
    $html .= '<td>';
    if    ($cfg->{identification}->{mode} eq "ase") {
        $html .= $cgi->start_form(-action => $ase->url(-action => "logout", -mode_during => "ase", -mode_after => "asecomeback"));
        $html .= "<div>" . $cgi->submit('submit','logout') . "</div>";
        $html .= $cgi->end_form;
    }
    elsif ($cfg->{identification}->{mode} eq "basicauth") {
        # nop
    }
    elsif ($cfg->{identification}->{mode} eq "naive") {
        $html .= $cgi->start_form(-action => "$myurl?page=logout");
        $html .= "<div>" . $cgi->submit('submit','logout') . "</div>";
        $html .= $cgi->end_form;
    }
    elsif ($cfg->{identification}->{mode} eq "constant") {
        # nop
    }
    else {
        # nop, misconfigured
    }
    $html .= '</td>';
    return $html;
}

sub viewlogout()
{
    my ($header, $cookie, $redirect);
    $header = {};
    $cookie = $cgi->cookie(
        -name      => 'username',
        -value     => ' ',
        -expires   => 1,
        -path      => '/'
    );
    $redirect = $cgi->redirect(
        -uri       => $myurl
    );

    $header = {
        -redirect => $redirect,
        -cookie => $cookie
    };
    return $cgi->redirect($header);
}

sub viewdropxmlform()
{
    my $html;
    $html = '';
    $html .= '<td>';
    $html .= $cgi->start_form(-action => "$myurl?page=dropxml");
    $html .= "<div>" . $cgi->submit('submit','dropxml') . "</div>";
    $html .= $cgi->end_form;
    $html .= '</td>';
    return $html;
}

sub execassociation()
{
    my $html;
    my ($formstruct, $username, $headerout);

    $html = '';

    $username = &identifyusername();
    if (not defined $username) {
        $html .= "<h2>Access&nbsp;denied</h2>";
        $html .= "Login to authenticate";
        return $html;
    }

    $formstruct = {};
    foreach my $k (keys %{$cgi->Vars()}) {
        if ($k =~ m|^([^/]+)/([^/]+)/([^/]+)$|) {
            $formstruct->{$1}->{$2}->{$3} = $cgi->param($k);
            $cgi->delete(-name=>$k);
        }
    }

    $headerout = 0;
    foreach my $mode ("arrival", "active", "departure") {
        if ($cgi->param("$mode/modify")) {
            foreach my $k (keys %{$formstruct->{$mode}}) {
                if ($formstruct->{$mode}->{$k}->{check} eq 'on') {
                    $html .= "<h3>Updating</h3>" unless($headerout++);

                    my $associate;
                    $associate = '';
                    $associate = "associated = 'true', " if ($mode eq "arrival");
                    $dbh->{AutoCommit} = 1;
                    $sql = sprintf("UPDATE reg_instance SET %s registry_desc = ? WHERE ( uuid_registry = ? );", $associate);
                    $sth = $dbh->prepare($sql);
                    $rv = $sth->execute($formstruct->{$mode}->{$k}->{registry_desc}, $k);
                    if (not defined $rv) {
                        $msg = $dbh->errstr;
                        $msg =~ s/[ ]*ERROR:?[ ]*//;
                        $html .= sprintf("<br/><img src=\"?page=gif;name=icon-x\">&nbsp;<code><b>%s</b></code> ERROR DataBase reports %s\n", $k, $msg);
                        next;
                    }
                    if ($rv != 1) {
                        $html .= sprintf("<br/><img src=\"?page=gif;name=icon-x\">&nbsp;<code><b>%s</b></code> ERROR update failed rv=$rv\n", $k);
                        next;
                    }
                    $html .= sprintf("<br/><img src=\"?page=gif;name=icon-ok\">&nbsp<code><b>%s</b></code> %s\n", $k, $formstruct->{$mode}->{$k}->{registry_desc});
                }
            }
        }
    }
    $html .= "<hr>\n" if($headerout);

    $headerout = 0;
    foreach my $mode ("arrival", "active", "departure") {
        if ($cgi->param("$mode/delete")) {
            foreach my $k (keys %{$formstruct->{$mode}}) {
                if ($formstruct->{$mode}->{$k}->{check} eq 'on') {
                    $html .= "<h3>Deleting</h3>" unless($headerout++);

                    $dbh->{AutoCommit} = 1;
                    $sql = sprintf("DELETE FROM reg_instance WHERE ( uuid_registry = ? );");
                    $sth = $dbh->prepare($sql);
                    $rv = $sth->execute($k);
                    if (not defined $rv) {
                        $msg = $dbh->errstr;
                        $msg =~ s/[ ]*ERROR:?[ ]*//;
                        $html .= sprintf("<br/><img src=\"?page=gif;name=icon-x\">&nbsp;<code><b>%s</b></code> ERROR DataBase reports %s\n", $k, $msg);
                        next;
                    }
                    if ($rv != 1) {
                        $html .= sprintf("<br/><img src=\"?page=gif;name=icon-x\">&nbsp;<code><b>%s</b></code> ERROR deletion failed rv=$rv\n", $k);
                        next;
                    }
                    $html .= sprintf("<br/><img src=\"?page=gif;name=icon-ok\">&nbsp<code><b>%s</b></code> %s\n", $k, $formstruct->{$mode}->{$k}->{registry_desc});
                }
            }
        }
    }
    $html .= "<hr>\n" if($headerout);

    return $html;
}

sub viewassociation()
{
    my $html;
    my $username;

    $html = '';
    $username = &identifyusername();
    if (not defined $username) {
        $html .= "<h2>Access&nbsp;denied</h2>";
        $html .= "Login to authenticate";
        return $html;
    }
    $html .= sprintf("<h2>Instances related to %s</h2>", $username);
    $html .= "<hr>\n";

    $html .= &execassociation();
    $html .= &condassociation("arrival", $username);
    $html .= &condassociation("active", $username);
    $html .= &condassociation("departure", $username);

    return $html;
}

sub condassociation($)
{
    my ($mode, $username) = @_;
    my ($html, $where, $headline);
    $html = '';

    if    ($mode eq "arrival") {
        $where = "associated = 'false'";
        $headline = "Arrival - queue of registered instances waiting for association";
    }
    elsif ($mode eq "active") {
        $where = "associated = 'true' AND registry_date >= (now() - interval '2 days')";
        $headline = "Active - associated instances";
    }
    elsif ($mode eq "departure") {
        $where = "associated = 'true' AND registry_date < (now() - interval '2 days')";
        $headline = "Departure - associated instances without heartbeat";
    }
    else {
        die;
    }

    $html .= "<table class=\"association\">\n";
    $html .= "<tr>";
    $html .= "<td colspan=\"4\">";
    $html .= $cgi->start_form(-action => "$myurl?page=association");
    $html .= "<h3>$headline</h3>";
    $html .= "</td>";
    $html .= "</tr>\n";

    $sql = sprintf("SELECT * FROM reg_instance WHERE ( registry_user = '%s' AND %s );", $username, $where);
    $rv = $dbh->selectall_hashref($sql, "uuid_registry");
    if (not $rv) {
        die "ERROR:$0: DataBase error: ".$dbh->errstr."\n";
    }
    if ((keys %{$rv}) == "") {
        $html .= "<tr>\n";
        $html .= "<td>---empty---</td>";
        $html .= "<td></td>";
        $html .= "<td></td>";
        $html .= "<td></td>";
        $html .= "</tr>\n";
    }
    else {
        $html .= "<tr>";

        #   checkallboxes
        $html .= "<td>";
        $html .= $cgi->checkbox(
            -name     => "$mode/checkallboxes",
            -checked  => ($mode eq "arrival" ? 1 : 0),
            -label    => "",
            -onClick  => "checkallboxes(this)"
        ) . "</td>";

        $html .= "<td><b>date</b></td>";
        $html .= "<td><b>registry</b></td>";
        $html .= "<td><b>description</b></td>";
        $html .= "</tr>\n";
        for my $i (sort keys %{$rv}) {
            $html .= "<tr>";

            #   checkbox
            $html .= "<td>";
            $html .= $cgi->checkbox(
                -name     => "$mode/$i/check",
                -checked  => ($mode eq "arrival" ? 1 : 0),
                -label    => ""
            ) . "</td>";

            #   date
            $html .= "<td>" . $rv->{$i}->{registry_date} . "</td>";

            #   registry
            $html .= "<td><code><b>" . $i . "</b></code></td>";

            #   description
            $html .= "<td>";
            $html .= $cgi->textfield(
                -name      => "$mode/$i/registry_desc",
                -override  => 1,
                -size      => 50,
                -maxlength => 50,
                -default   => $rv->{$i}->{registry_desc},
            ) . "</td>";

        $html .= "</tr>\n";
        }

        $html .= "<tr>";
        $html .= "<td colspan=\"2\">";
        $cgi->delete(-name=>'username');
        $html .= "<div>" . $cgi->hidden(-name=>'username', -value => $username) . "</div>";
        $html .= "<div>" . $cgi->submit("$mode/delete",'delete') . "</div>";
        $html .= "</td>";
        $html .= "<td colspan=\"2\">";
        if ($mode eq "arrival") {
            $html .= $cgi->submit("$mode/modify",'modify and associate')
        }
        else {
            $html .= $cgi->submit("$mode/modify",'modify and refresh')
        }
        $html .= "</td>";
        $html .= "</tr>\n";
    }
    $html .= $cgi->end_form;
    $html .= "</table>";
}

sub execdropxml()
{
    my $html;
    my $data;

    $html = '';
    if (&uao()) {
        $html .= "<?xml version=\"1.0\" encoding=\"iso-8859-1\" standalone=\"no\"?>\n";
        $html .= "<!DOCTYPE registry\n";
        $html .= "  PUBLIC \"-//OpenPKG//DTD OpenPKG Registry 0.0.1//EN\"\n";
        $html .= "  \"http://registry.openpkg.org/registry.dtd\" []>\n";
    }

    $data = $cgi->param("data");
    if (not defined $data) {
        if (&uao()) {
            $html .= "<registry>\n";
            $html .= "</registry>\n";
        }
        return($html)
    }

    $data =~ s|^[^<]*<|<|s; # cut off crap preceeding opening brace
    $data =~ s|>[^>]*$|>|s; # cut off crap succeeding closing brace
    if ($data eq "") {
        if (&uao()) {
            $html .= "<registry>\n";
            $html .= "</registry>\n";
        }
        return($html)
    }

    #   XMLin() requires the string being passed to start with '<'
    #   otherwise it is assumed the parameter is a filename to be
    #   read. We check for the closing brace at the end, too.
    #
    if (not $data =~ m|^<.*>$|s) {
        if (&uao()) {
            $html .= "<registry>\n";
            $html .= "ERROR: not XML data\n";
            $html .= "</registry>\n";
        }
        else {
            $html .= sprintf("<br/><img src=\"?page=gif;name=icon-x\">&nbsp;ERROR: not XML data\n");
        }
        return $html;
    }

    #   Check DTD
    #   FIXME currently we just drop it!
    $data =~ s|<!DOCTYPE registry[^>]*>||s;

    my $ref = eval { local $SIG{__DIE__}; XMLin($data, ForceArray => 1) };
    if ($@) {
        $msg = $@;
        $msg =~ s| at \/.*$||; # hide program code file name and line
        if (&uao()) {
            $html .= "<registry>\n";
            $html .= sprintf("ERROR: XML parser reports %s\n", $msg);
            $html .= "</registry>\n";
        }
        else {
            $html .= sprintf("<br/><img src=\"?page=gif;name=icon-x\">&nbsp;ERROR: XML parser reports %s\n", $msg);
        }
        return $html;
    }

    #   begin the transaction
    #
    $dbh->{AutoCommit} = 0;
    eval { $rc  = $dbh->begin_work };
    if ($@) {
        $msg = $@;
        $msg =~ s/[ ]*ERROR:?[ ]*//;
        if (&uao()) {
            $html .= "<registry>\n";
            $html .= sprintf("ERROR: at begin of transaction, DataBase reports %s\n", $msg);
            $html .= "</registry>\n";
        }
        else {
            $html .= sprintf("<br/><img src=\"?page=gif;name=icon-x\">&nbsp;ERROR: at begin of transaction, DataBase reports %s\n", $msg);
        }
        return $html;
    }
    my $commit = 1;

    if (&uao()) {
        $html .= "<registry>\n";
    }
    foreach my $k (keys %{$ref->{request}}) {
        my $username;
        $username = $ref->{request}->{$k}->{registry_user};

        #   check for known username by updating heartbeat and look for database errors
        #
        $sql = sprintf("UPDATE reg_user SET heartbeat = now() WHERE ( username = '%s' );", $username);
        $rv = $dbh->do($sql);
        if (not defined $rv) {
            $msg = $dbh->errstr;
            $msg =~ s/[ ]*ERROR:?[ ]*//;
            if (&uao()) {
                $html .= sprintf("    <response id=\"%s\" done=\"no\">ERROR DataBase reports %s</response>\n", $k, $msg);
            }
            else {
                $html .= sprintf("<br/><img src=\"?page=gif;name=icon-x\">&nbsp;<b>%s</b> ERROR DataBase reports %s\n", $k, $msg);
            }
            $commit = 0;
            next;
        }
        if ($rv != 1) {
            if (&uao()) {
                $html .= sprintf("    <response id=\"%s\" done=\"no\">ERROR username %s not found</response>\n", $k, $username);
            }
            else {
                $html .= sprintf("<br/><img src=\"?page=gif;name=icon-x\">&nbsp;<b>%s</b> ERROR username <i>%s</i> not found\n", $k, $username);
            }
            $commit = 0;
            next;
        }

        #   insert instance, possibly fall back to update
        #
        my ($fields, $fieldlist, $fieldbind, $fieldvals, $sth);
        ($sql, $fields, $fieldlist, $fieldbind, $fieldvals, $sth) = undef;
        for my $field (sort keys %{$ref->{request}->{$k}}) {
            $fieldlist = (defined $fieldlist ? $fieldlist . ", " : "") . $field;
            $fieldbind = (defined $fieldbind ? $fieldbind . ", " : "") . "?";
            push @{$fieldvals}, $ref->{request}->{$k}->{$field};
            #FIXME here we could check the validity of all fields with string 'uuid' in their names
        }
        $sql = sprintf("INSERT INTO reg_instance (%s) VALUES (%s);", $fieldlist, $fieldbind);
        $sth = $dbh->prepare($sql);
        $rv = $sth->execute(@{$fieldvals});
        if (not defined $rv) {
            $msg = $dbh->errstr;
            $msg =~ s/[ ]*ERROR:?[ ]*//;
            unless ($msg =~ m|duplicate key violates unique constraint|) {
                if (&uao()) {
                    $html .= sprintf("    <response id=\"%s\" done=\"no\">ERROR DataBase reports %s</response>\n", $k, $msg);
                }
                else {
                    $html .= sprintf("<br/><img src=\"?page=gif;name=icon-x\">&nbsp;<b>%s</b> ERROR DataBase reports %s\n", $k, $msg);
                }
                $commit = 0;
                next;
            }
        }
        if ($rv != 1) {
            my $rowkey;
            $rowkey = undef;
            ($sql, $fields, $fieldlist, $fieldbind, $fieldvals, $sth) = undef;
            for my $field (sort keys %{$ref->{request}->{$k}}) {
                if ($field eq "uuid_registry") {
                    $rowkey = $ref->{request}->{$k}->{$field}
                }
                else {
                    $fieldlist = (defined $fieldlist ? $fieldlist . ", " : "") . $field . " = ?";
                    push @{$fieldvals}, $ref->{request}->{$k}->{$field};
                }
            }
            $sql = sprintf("UPDATE reg_instance SET %s WHERE ( uuid_registry = ? );", $fieldlist);
            $sth = $dbh->prepare($sql);
            $rv = $sth->execute(@{$fieldvals}, $rowkey);
            if (not defined $rv) {
                $msg = $dbh->errstr;
                $msg =~ s/[ ]*ERROR:?[ ]*//;
                if (&uao()) {
                    $html .= sprintf("    <response id=\"%s\" done=\"no\">ERROR DataBase reports %s</response>\n", $k, $msg);
                }
                else {
                    $html .= sprintf("<br/><img src=\"?page=gif;name=icon-x\">&nbsp;<b>%s</b> ERROR DataBase reports %s\n", $k, $msg);
                }
                $commit = 0;
                next;
            }
            if ($rv != 1) {
                if (&uao()) {
                    $html .= sprintf("    <response id=\"%s\" done=\"no\">ERROR update failed rv=%s</response>\n", $k, $msg);
                }
                else {
                    $html .= sprintf("<br/><img src=\"?page=gif;name=icon-x\">&nbsp;<b>%s</b> ERROR update failed rv=$rv\n", $k);
                }
                $commit = 0;
                next;
            }
        }
        if (&uao()) {
            $html .= sprintf("    <response id=\"%s\" done=\"yes\">%s</response>\n", $k, $ref->{request}->{$k}->{registry_desc});
        }
        else {
            $html .= sprintf("<br/><img src=\"?page=gif;name=icon-ok\">&nbsp<b>%s</b> %s\n", $k, $ref->{request}->{$k}->{registry_desc});
        }
    }

    if ($commit == 1) {
        eval { $dbh->commit };
        if ($@) {
            $msg = $@;
            $msg =~ s/[ ]*ERROR:?[ ]*//;
            if (&uao()) {
                $html .= sprintf("    <response id=\"FIXME\" done=\"no\">ERROR commit transaction failed, DataBase reports %s</response>\n", $msg);
            }
            else {
                $html .= sprintf("<br/><img src=\"?page=gif;name=icon-x\">&nbsp;ERROR commit transaction failed, DataBase reports %s\n", $msg);
            }
        }
        $cgi->delete(-name=>'data');
    }
    else {
        eval { $dbh->rollback };
        if ($@) {
            $msg = $@;
            $msg =~ s/[ ]*ERROR:?[ ]*//;
            if (&uao()) {
                $html .= sprintf("    <response id=\"FIXME\" done=\"no\">ERROR rollback transaction failed, DataBase reports %s</response>\n", $msg);
            }
            else {
                $html .= sprintf("<br/><img src=\"?page=gif;name=icon-x\">&nbsp;ERROR rollback transaction failed, DataBase reports %s\n", $msg);
            }
        }
    }
    if (&uao()) {
        $html .= "</registry>\n";
    }
    return($html);
}

sub viewdropxml()
{
    my $html;
    my $data;
    $html = '';

    $html .= &execdropxml();
    return $html if (&uao());

    if (defined $cgi->param("data")) {
        $html .= "<h2>Correct registration data below</h2>";
    }
    else {
        $html .= "<h2>Paste registration data below</h2>";
    }

    $html .= "      <table class=\"menu\">\n";
    $html .= "          <tr>\n";
    $html .= "              <td>\n";
    $html .= $cgi->start_form(-action => "$myurl?page=dropxml");
    $html .= "<div>" . $cgi->textarea(
        -name       => 'data',
        -columns    => 80,
        -rows       => 15,
        -default    => '',
    ) . "</div>\n";
    $html .= "<div>" . $cgi->submit('submit','register') . "</div>";
    $html .= $cgi->end_form;
    $html .= "              </td>\n";
    $html .= "          </tr>\n";
    $html .= "      </table>\n";
    return $html;
}

sub printjscheckallboxes()
{
    my $js;
    $js = <<'EOT';
<script type="text/javascript"><!--
function checkallboxes(me) {
    var value=me.checked;
    var form=me.form;
    var items=form.length;
    var i;
    for(i=0; i<items; i++) {
        if(form[i].type=='checkbox') {
            form[i].checked=value;
        }
    }
}
//--></script>
EOT
    return $js;
}

sub canvas($)
{
    my ($page) = @_;
    my ($http, $head, $body, $canvas);

    $http = $page;
    $http =~ s|<html>.*$||s;

    $head = $page;
    $head =~ s|^.*<head[^>]*>||s;
    $head =~ s|</head>.*||s;

    $body = $page;
    $body =~ s|^.*<body[^>]*>||s;
    $body =~ s|</body>.*||s;

    (undef, undef, $canvas) = &fetchurlcached($cfg->{canvas}->{url});
    if (not defined $canvas or $canvas eq "") {
        $canvas =
            "<!DOCTYPE html PUBLIC \"-//W3C//DTD XHTML 1.0 Strict//EN\"\n" .
            "    \"http://www.w3.org/TR/xhtml1/DTD/xhtml1-strict.dtd\">\n" .
            "<html>\n" .
            "    <head>\n" .
            "    <title>OpenPKG registry</title>\n" .
                     $cfg->{canvas}->{mark_head} . "\n" .
            "    </head>\n" .
            "    <body class=\"registry\">\n" .
                     $cfg->{canvas}->{mark_body} . "\n" .
            "    </body>\n" .
            "</html>\n";
    }

    $canvas =~ s|$cfg->{canvas}->{mark_head}|$head|;
    $canvas =~ s|$cfg->{canvas}->{mark_body}|$body|;
    return $http . $canvas;
}

sub fetchurlcached ($)
{
    my ($url) = @_;
    my ($content_type, $expires, $content);
    undef $content;

    ($content_type, $expires, $content) = &getcache($url) if ($cfg->{canvas}->{urlcache});
    ($content_type, $expires, $content) = &fetchurl($url) if (not defined $content);
    &setcache($url, $content_type, $expires, $content) if (defined $content and $cfg->{canvas}->{urlcache});
    return $content_type, $expires, $content;
}

sub getcache ($)
{
    my ($url) = @_;
    my ($content_type, $expires, $content, $rv, $sth);
    ($content_type, $expires, $content) = undef;

    #   invalidate expired records
    $rv = $dbs->do("DELETE FROM cache WHERE ( expires <= ? );", undef, time()) or die $dbs->errstr(); #FIXME

    #   dig in the cache
    $sth = $dbs->prepare_cached("SELECT content_type, expires, content FROM cache WHERE url = ?;") or die $dbs->errstr(); #FIXME
    $sth->execute($url) or die $dbs->errstr(); #FIXME
    $rv = $sth->fetchrow_hashref;

    $content_type = $rv->{content_type};
    $expires      = $rv->{expires};
    $content      = $rv->{content};
    return $content_type, $expires, $content;
}

sub setcache ($$$$)
{
    my ($url, $content_type, $expires, $content) = @_;

    $expires = time() + 600 if (not defined $expires);

    $rv = $dbs->do("INSERT INTO cache (url, content_type, expires, content) VALUES (?, ?, ?, ?);",
                   undef,
                   $url, $content_type, $expires, $content) or die $dbs->errstr(); #FIXME
}

sub fetchurl ($)
{
    my ($url) = @_;
    my ($content_type, $expires, $content, $rv, $sth);
    ($content_type, $expires, $content) = undef;

    my $response;
    use HTTP::Response;
    use Socket;
    use Net::HTTP;
    use LWP::UserAgent;
    my $ua = new LWP::UserAgent;
    $ua->agent("openpkg-$progname/$progvers");
    $ua->timeout(20);
    $ua->max_size(1*1024*1024);
    $ua->max_redirect(2);
    $ua->protocols_allowed([ 'http', 'https']);
    $response = $ua->get($url);
    if ($response->is_success) {
        $content_type = $response->content_type;
        $expires      = $response->expires;
        $content      = $response->content;
    }

    return $content_type, $expires, $content;
}

sub identifyusername()
{
    my $username;
    $username = undef;

    if    ($cfg->{identification}->{mode} eq "ase") {
        $username = $ase->attr("client-login-name") if ($ase->login);
    }
    elsif ($cfg->{identification}->{mode} eq "basicauth") {
        $username = $ENV{'REMOTE_USER'};
    }
    elsif ($cfg->{identification}->{mode} eq "naive") {
        $username = $cgi->cookie("username");
    }
    elsif ($cfg->{identification}->{mode} eq "constant") {
        $username = $cfg->{identification}->{default};
    }
    else {
        # nop, misconfigured
    }
    $username = undef if ($username =~ m|^[ ]+$|);
    return $username;
}

